from unittest.mock import Mock
import unittest

from apipeline.pipeline.pipeline import Pipeline
from apipeline.processors.frame_processor import FrameProcessor


class TestMock(unittest.TestCase):
    class MockProcessor(FrameProcessor):
        def __init__(self, name):
            self.name = name

        def __str__(self):
            return self.name

    def setUp(self):
        self.processor1 = self.MockProcessor("processor1")
        self.processor2 = self.MockProcessor("processor2")
        self.pipeline = Pipeline([self.processor1, self.processor2])
        self.pipeline.name = "MyClass"


class TestMyFunction(unittest.TestCase):
    def test_my_function(self):
        external_service = Mock()
        external_service.get_data.return_value = {"key": "value"}

        result = my_function(external_service)

        self.assertEqual(result, "value")
        external_service.get_data.assert_called_once()


def my_function(service):
    data = service.get_data()
    return data["key"]


if __name__ == "__main__":
    unittest.main()



from typing import Awaitable, Callable

from apipeline.frames.sys_frames import Frame, SystemFrame
from apipeline.processors.frame_processor import FrameDirection, FrameProcessor


class FunctionFilter(FrameProcessor):
    def __init__(self, filter: Callable[[Frame], Awaitable[bool]]):
        super().__init__()
        self._filter = filter

    #
    # Frame processor
    #

    def _should_passthrough_frame(self, frame):
        return isinstance(frame, SystemFrame)

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        passthrough = self._should_passthrough_frame(frame)
        allowed = await self._filter(frame)
        if passthrough or allowed:
            await self.push_frame(frame, direction)



from typing import List, Optional

from apipeline.frames.sys_frames import Frame
from apipeline.processors.frame_processor import FrameDirection, FrameProcessor


class FrameFilter(FrameProcessor):
    def __init__(
        self, ignored_frame_types: Optional[list] = [], include_frame_types: Optional[list] = None
    ):
        super().__init__()
        self._ignored_frame_types = (
            tuple(ignored_frame_types) if ignored_frame_types is not None else None
        )
        self._include_frame_types = tuple(include_frame_types) if include_frame_types else None

    def _should_passthrough_frame(self, frame):
        if self._ignored_frame_types is not None and not isinstance(
            frame, self._ignored_frame_types
        ):
            if not self._include_frame_types or (
                self._include_frame_types and isinstance(frame, self._include_frame_types)
            ):
                return False
        return True

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        if self._should_passthrough_frame(frame):
            await self.push_frame(frame, direction)



#
# Copyright (c) 2024, Daily
#
# SPDX-License-Identifier: BSD 2-Clause License
#

from enum import Enum
import asyncio
import logging
import time


from apipeline.frames.sys_frames import ErrorFrame, Frame, MetricsFrame, StartInterruptionFrame
from apipeline.frames.control_frames import StartFrame
from apipeline.utils.obj import obj_count, obj_id


class FrameDirection(Enum):
    DOWNSTREAM = 1
    UPSTREAM = 2


class FrameProcessorMetrics:
    def __init__(self, name: str):
        self._name = name
        self._start_ttfb_time = 0
        self._start_processing_time = 0
        self._should_report_ttfb = True

    async def start_ttfb_metrics(self, report_only_initial_ttfb):
        if self._should_report_ttfb:
            self._start_ttfb_time = time.time()
            self._should_report_ttfb = not report_only_initial_ttfb

    async def stop_ttfb_metrics(self):
        if self._start_ttfb_time == 0:
            return None

        value = time.time() - self._start_ttfb_time
        logging.debug(f"{self._name} TTFB: {value}")
        ttfb = {"processor": self._name, "value": value}
        self._start_ttfb_time = 0
        return MetricsFrame(ttfb=[ttfb])

    async def start_processing_metrics(self):
        self._start_processing_time = time.time()

    async def stop_processing_metrics(self):
        if self._start_processing_time == 0:
            return None

        value = time.time() - self._start_processing_time
        logging.debug(f"{self._name} processing time: {value}")
        processing = {"processor": self._name, "value": value}
        self._start_processing_time = 0
        return MetricsFrame(processing=[processing])


class FrameProcessor:
    def __init__(
        self, *, name: str | None = None, loop: asyncio.AbstractEventLoop | None = None, **kwargs
    ):
        self.id: int = obj_id()
        self.name = name or f"{self.__class__.__name__}#{obj_count(self)}"
        self._parent_pipeline: "FrameProcessor" | None = None
        self._prev: "FrameProcessor" | None = None
        self._next: "FrameProcessor" | None = None
        self._loop: asyncio.AbstractEventLoop = loop or asyncio.get_running_loop()

        # Properties
        self._allow_interruptions = False
        self._enable_metrics = False
        self._enable_usage_metrics = False
        self._report_only_initial_ttfb = False

        # Metrics
        self._metrics = FrameProcessorMetrics(name=self.name)

        # SkipFrames
        self._skip_frames = []

    @property
    def interruptions_allowed(self):
        return self._allow_interruptions

    @property
    def metrics_enabled(self):
        return self._enable_metrics

    @property
    def usage_metrics_enabled(self):
        return self._enable_usage_metrics

    @property
    def report_only_initial_ttfb(self):
        return self._report_only_initial_ttfb

    def add_skip_frame(self, frame: Frame):
        self._skip_frames.append(frame)

    def can_generate_metrics(self) -> bool:
        return False

    async def start_ttfb_metrics(self):
        if self.can_generate_metrics() and self.metrics_enabled:
            await self._metrics.start_ttfb_metrics(self._report_only_initial_ttfb)

    async def stop_ttfb_metrics(self):
        if self.can_generate_metrics() and self.metrics_enabled:
            frame = await self._metrics.stop_ttfb_metrics()
            if frame:
                await self.push_frame(frame)

    async def start_processing_metrics(self):
        if self.can_generate_metrics() and self.metrics_enabled:
            await self._metrics.start_processing_metrics()

    async def stop_processing_metrics(self):
        if self.can_generate_metrics() and self.metrics_enabled:
            frame = await self._metrics.stop_processing_metrics()
            if frame:
                await self.push_frame(frame)

    async def stop_all_metrics(self):
        await self.stop_ttfb_metrics()
        await self.stop_processing_metrics()

    async def cleanup(self):
        pass

    def link(self, processor: "FrameProcessor"):
        self._next = processor
        processor._prev = self
        logging.debug(f"Linking {self} -> {self._next}")

    def get_event_loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    def set_parent_pipeline(self, pipeline: "FrameProcessor"):
        self._parent_pipeline = pipeline

    def get_parent_pipeline(self) -> "FrameProcessor":
        return self._parent_pipeline

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        if frame in self._skip_frames:
            return

        if isinstance(frame, StartFrame):
            self._allow_interruptions = frame.allow_interruptions
            self._enable_metrics = frame.enable_metrics
            self._enable_usage_metrics = frame.enable_usage_metrics
            self._report_only_initial_ttfb = frame.report_only_initial_ttfb
        elif isinstance(frame, StartInterruptionFrame):
            await self.stop_all_metrics()

    async def push_error(self, error: ErrorFrame):
        await self.push_frame(error, FrameDirection.UPSTREAM)

    async def push_frame(self, frame: Frame, direction: FrameDirection = FrameDirection.DOWNSTREAM):
        try:
            if direction == FrameDirection.DOWNSTREAM and self._next:
                if hasattr(logging, "trace"):
                    logging.trace(f"Pushing {frame} from {self} to {self._next}")
                else:
                    logging.debug(f"Pushing {frame} from {self} to {self._next}")
                await self._next.process_frame(frame, direction)
            elif direction == FrameDirection.UPSTREAM and self._prev:
                if hasattr(logging, "trace"):
                    logging.trace(f"Pushing {frame} from {self} to {self._next}")
                else:
                    logging.debug(f"Pushing {frame} upstream from {self} to {self._prev}")
                await self._prev.process_frame(frame, direction)
        except Exception as e:
            logging.exception(f"Uncaught exception in {self}: {e}")

    def __str__(self):
        return self.name



import re


ENDOFSENTENCE_PATTERN_STR = r"""
    (?<![A-Z])       # Negative lookbehind: not preceded by an uppercase letter (e.g., "U.S.A.")
    (?<!\d)          # Negative lookbehind: not preceded by a digit (e.g., "1. Let's start")
    (?<!\d\s[ap])    # Negative lookbehind: not preceded by time (e.g., "3:00 a.m.")
    (?<!Mr|Ms|Dr)    # Negative lookbehind: not preceded by Mr, Ms, Dr (combined bc. length is the same)
    (?<!Mrs)         # Negative lookbehind: not preceded by "Mrs"
    (?<!Prof)        # Negative lookbehind: not preceded by "Prof"
    [\.。\?？\!！:：] # Match a period, question mark, exclamation point, or colon
    $                # End of string
"""
ENDOFSENTENCE_PATTERN = re.compile(ENDOFSENTENCE_PATTERN_STR, re.VERBOSE)


def match_endofsentence(text: str) -> bool:
    return ENDOFSENTENCE_PATTERN.search(text.rstrip()) is not None



import logging
import asyncio

import os
import unittest

from apipeline.frames import Frame, ImageRawFrame, DataFrame, TextFrame, EndFrame
from apipeline.pipeline.pipeline import Pipeline
from apipeline.pipeline.runner import PipelineRunner
from apipeline.pipeline.task import PipelineParams, PipelineTask
from apipeline.processors.aggregators.gated import GatedAggregator
from apipeline.processors.output_processor import OutputProcessor


"""
python -m unittest tests.aggr.test_gated_aggr.TestGatedAggregator
"""


class PrintOutputProcessor(OutputProcessor):
    async def sink(self, frame: DataFrame):
        print(f"frame--->:{frame}")
        self._sink_event.set()


class TestGatedAggregator(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls):
        logging.basicConfig(
            level=os.getenv("LOG_LEVEL", "info").upper(),
            format="%(asctime)s - %(name)s - %(levelname)s - %(pathname)s:%(lineno)d - %(funcName)s - %(message)s",
            handlers=[logging.StreamHandler()],
        )
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    async def asyncSetUp(self):
        pass

    async def asyncTearDown(self):
        pass

    async def test_gated_print(self):
        aggregator = GatedAggregator(
            gate_close_fn=lambda x: isinstance(x, TextFrame),
            gate_open_fn=lambda x: isinstance(x, ImageRawFrame),
            init_start_open=False,
        )
        pipeline = Pipeline([aggregator, PrintOutputProcessor()])
        task = PipelineTask(pipeline, PipelineParams())

        await task.queue_frame(TextFrame("Hello, "))
        await task.queue_frame(TextFrame("Hello again."))
        await task.queue_frame(
            ImageRawFrame(
                image=bytes([]),
                size=(0, 0),
                format="JPEG",
                mode="RGB",
            )
        )
        await task.queue_frame(TextFrame("Goodbye1."))
        await task.queue_frame(TextFrame("Goodbye2."))
        await task.queue_frame(
            ImageRawFrame(
                image=bytes([]),
                size=(0, 0),
                format="JPEG",
                mode="RGB",
            )
        )
        await task.queue_frame(TextFrame("Goodbye3."))
        await task.queue_frame(EndFrame())

        runner = PipelineRunner()
        await asyncio.gather(runner.run(task))



from typing import List
import logging

from apipeline.frames.data_frames import DataFrame
from apipeline.frames.control_frames import StartFrame
from apipeline.processors.frame_processor import FrameProcessor
from apipeline.processors.input_processor import InputProcessor
from apipeline.processors.output_processor import OutputProcessor


class TestException(BaseException):
    pass


class TestFrameProcessor(FrameProcessor):
    def __init__(self, test_frames):
        self.test_frames = test_frames
        self._list_counter = 0
        super().__init__()

    async def process_frame(self, frame, direction):
        await super().process_frame(frame, direction)

        if not self.test_frames[
            0
        ]:  # then we've run out of required frames but the generator is still going?
            raise TestException(f"Oops, got an extra frame, {frame}")
        if isinstance(self.test_frames[0], List):
            # We need to consume frames until we see the next frame type after this
            next_frame = self.test_frames[1]
            if isinstance(frame, next_frame):
                # we're done iterating the list I guess
                print(f"TestFrameProcessor got expected list exit frame: {frame}")
                # pop twice to get rid of the list, as well as the next frame
                self.test_frames.pop(0)
                self.test_frames.pop(0)
                self.list_counter = 0
            else:
                fl = self.test_frames[0]
                fl_el = fl[self._list_counter % len(fl)]
                if isinstance(frame, fl_el):
                    print(f"TestFrameProcessor got expected list frame: {frame}")
                    self._list_counter += 1
                else:
                    raise TestException(f"Inside a list, expected {fl_el} but got {frame}")

        else:
            if not isinstance(frame, self.test_frames[0]):
                raise TestException(f"Expected {self.test_frames[0]}, but got {frame}")
            print(f"TestFrameProcessor got expected frame: {frame}")
            self.test_frames.pop(0)


class TestInputPorcessor(InputProcessor):
    async def start(self, frame: StartFrame):
        logging.debug(f"start input frame: {frame}")

    async def stop(self):
        logging.debug(f"stop input frame")


class TestOutputPorcessor(OutputProcessor):
    async def start(self, frame: StartFrame):
        super().start()
        logging.debug(f"start output frame: {frame}")

    async def stop(self):
        super().stop()
        logging.debug(f"stop output frame")

    async def sink(self, frame: DataFrame):
        logging.debug(f"sink data frame")



import logging
import asyncio

import os
import unittest

from apipeline.frames import Frame, ImageRawFrame, DataFrame, TextFrame, EndFrame, StopTaskFrame
from apipeline.frames.control_frames import ControlFrame, SyncNotifyFrame
from apipeline.frames.sys_frames import CancelFrame, StartInterruptionFrame, StopInterruptionFrame
from apipeline.notifiers.event_notifier import EventNotifier
from apipeline.pipeline.pipeline import Pipeline
from apipeline.pipeline.runner import PipelineRunner
from apipeline.pipeline.task import PipelineParams, PipelineTask
from apipeline.processors.aggregators.hold import HoldFramesAggregator, HoldLastFrameAggregator
from apipeline.processors.filters.function_filter import FunctionFilter
from apipeline.processors.frame_processor import FrameDirection, FrameProcessor
from apipeline.processors.user_idle_processor import UserIdleProcessor


"""
python -m unittest tests.aggr.test_hold_aggr.TestAggregator.test_hold_frames_print
python -m unittest tests.aggr.test_hold_aggr.TestAggregator.test_hold_frame_print
"""


class PrintOutFrameProcessor(FrameProcessor):
    def __init__(self):
        super().__init__()

    async def process_frame(self, frame, direction):
        await super().process_frame(frame, direction)
        logging.info(f"{frame}")
        if isinstance(frame, TextFrame) and frame.text == "end":
            await self.push_frame(EndFrame(), FrameDirection.UPSTREAM)
            logging.info("End ok")
        if isinstance(frame, TextFrame) and frame.text == "cancel":
            await self.push_frame(CancelFrame(), FrameDirection.UPSTREAM)
            logging.info("Cancel ok")
        if isinstance(frame, TextFrame) and frame.text == "endTask":
            await self.push_frame(StopTaskFrame(), FrameDirection.UPSTREAM)
            logging.info("End Task ok")


class TestAggregator(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls):
        logging.basicConfig(
            level=os.getenv("LOG_LEVEL", "info").upper(),
            format="%(asctime)s - %(name)s - %(levelname)s - %(pathname)s:%(lineno)d - %(funcName)s - %(message)s",
            handlers=[logging.StreamHandler()],
        )

    @classmethod
    def tearDownClass(cls):
        pass

    async def asyncSetUp(self):
        self._notifier = EventNotifier()

    async def asyncTearDown(self):
        pass

    async def wake_notifier_filter(self, frame: Frame):
        if isinstance(frame, SyncNotifyFrame):
            # await asyncio.sleep(1)
            await self._notifier.notify()
        return True

    async def test_hold_frames_print(self):
        async def idle_callback(processor: UserIdleProcessor):
            await self._notifier.notify()

        user_idle = UserIdleProcessor(callback=idle_callback, timeout=1.0)

        aggregator = HoldFramesAggregator(
            notifier=self._notifier,
            hold_frame_classes=(TextFrame,),
        )
        pipeline = Pipeline(
            [
                aggregator,
                # user_idle,
                FunctionFilter(filter=self.wake_notifier_filter),
                PrintOutFrameProcessor(),
            ]
        )
        task = PipelineTask(pipeline, PipelineParams())

        await task.queue_frame(TextFrame("Hello many frames, "))
        await task.queue_frame(SyncNotifyFrame())
        await task.queue_frame(
            ImageRawFrame(
                image=bytes([]),
                size=(0, 0),
                format="JPEG",
                mode="RGB",
            )
        )
        await task.queue_frame(TextFrame("Goodbye1."))
        await task.queue_frame(
            ImageRawFrame(
                image=bytes([]),
                size=(0, 0),
                format="PNG",
                mode="RGB",
            )
        )
        # await task.queue_frame(EndFrame())
        # await task.queue_frame(TextFrame("end"))
        await task.queue_frame(StopTaskFrame())

        runner = PipelineRunner()
        await asyncio.gather(runner.run(task))

    async def test_hold_frame_print(self):
        async def idle_callback(processor: UserIdleProcessor):
            await self._notifier.notify()

        user_idle = UserIdleProcessor(callback=idle_callback, timeout=1.0)

        aggregator = HoldLastFrameAggregator(
            notifier=self._notifier,
            hold_frame_classes=(TextFrame,),
        )
        pipeline = Pipeline(
            [
                aggregator,
                # user_idle,
                FunctionFilter(filter=self.wake_notifier_filter),
                PrintOutFrameProcessor(),
            ]
        )
        task = PipelineTask(pipeline, PipelineParams())

        await task.queue_frame(TextFrame("Hello last one frame, "))
        await task.queue_frame(SyncNotifyFrame())
        await task.queue_frame(
            ImageRawFrame(
                image=bytes([]),
                size=(0, 0),
                format="JPEG",
                mode="RGB",
            )
        )
        await task.queue_frame(TextFrame("Goodbye1."))
        await task.queue_frame(
            ImageRawFrame(
                image=bytes([]),
                size=(0, 0),
                format="PNG",
                mode="RGB",
            )
        )
        # await task.queue_frame(EndFrame())
        await task.queue_frame(TextFrame("end"))
        await task.queue_frame(TextFrame("endTask"))
        await task.queue_frame(StopTaskFrame())

        runner = PipelineRunner()
        await asyncio.gather(runner.run(task))



#
# Copyright (c) 2024, Daily
#
# SPDX-License-Identifier: BSD 2-Clause License
#

from itertools import chain
from typing import List
import logging
import asyncio

from apipeline.pipeline.base_pipeline import BasePipeline
from apipeline.pipeline.pipeline import Pipeline
from apipeline.processors.frame_processor import FrameDirection, FrameProcessor
from apipeline.frames.sys_frames import CancelFrame, Frame
from apipeline.frames.control_frames import EndFrame, StartFrame


class Source(FrameProcessor):
    def __init__(self, upstream_queue: asyncio.Queue):
        super().__init__()
        self._up_queue = upstream_queue

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        match direction:
            case FrameDirection.UPSTREAM:
                await self._up_queue.put(frame)
            case FrameDirection.DOWNSTREAM:
                await self.push_frame(frame, direction)


class Sink(FrameProcessor):
    def __init__(self, downstream_queue: asyncio.Queue):
        super().__init__()
        self._down_queue = downstream_queue

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        match direction:
            case FrameDirection.UPSTREAM:
                await self.push_frame(frame, direction)
            case FrameDirection.DOWNSTREAM:
                await self._down_queue.put(frame)


class ParallelPipeline(BasePipeline):
    def __init__(self, *args):
        super().__init__()

        if len(args) == 0:
            raise Exception(f"ParallelPipeline needs at least one argument")

        self._sources = []
        self._sinks = []

        self._up_queue = asyncio.Queue()
        self._down_queue = asyncio.Queue()
        self._up_task: asyncio.Task | None = None
        self._down_task: asyncio.Task | None = None

        self._pipelines = []

        logging.debug(f"Creating {self} pipelines")
        for processors in args:
            if not isinstance(processors, list):
                raise TypeError(f"ParallelPipeline argument {processors} is not a list")

            # We will add a source before the pipeline and a sink after.
            source = Source(self._up_queue)
            sink = Sink(self._down_queue)
            self._sources.append(source)
            self._sinks.append(sink)

            # Create pipeline
            pipeline = Pipeline(processors)
            source.link(pipeline)
            pipeline.link(sink)
            self._pipelines.append(pipeline)

        logging.debug(f"Finished creating {self} pipelines")

    #
    # BasePipeline
    #

    def processors_with_metrics(self) -> List[FrameProcessor]:
        return list(chain.from_iterable(p.processors_with_metrics() for p in self._pipelines))

    #
    # Frame processor
    #

    async def cleanup(self):
        await asyncio.gather(*[p.cleanup() for p in self._pipelines])

    async def _start_tasks(self):
        loop = self.get_event_loop()
        self._up_task = loop.create_task(self._process_up_queue())
        self._down_task = loop.create_task(self._process_down_queue())

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        if isinstance(frame, StartFrame):
            await self._start_tasks()

        if direction == FrameDirection.UPSTREAM:
            # If we get an upstream frame we process it in each sink.
            await asyncio.gather(*[s.process_frame(frame, direction) for s in self._sinks])
        elif direction == FrameDirection.DOWNSTREAM:
            # If we get a downstream frame we process it in each source.
            # TODO(aleix): We are creating task for each frame. For real-time
            # video/audio this might be too slow. We should use an already
            # created task instead.
            await asyncio.gather(*[s.process_frame(frame, direction) for s in self._sources])

        # If we get an EndFrame we stop our queue processing tasks and wait on
        # all the pipelines to finish.
        if isinstance(frame, CancelFrame) or isinstance(frame, EndFrame):
            # Use None to indicate when queues should be done processing.
            await self._up_queue.put(None)
            await self._down_queue.put(None)
            if self._up_task:
                await self._up_task
            if self._down_task:
                await self._down_task

    async def _process_up_queue(self):
        running = True
        seen_ids = set()
        while running:
            frame = await self._up_queue.get()
            if frame and frame.id not in seen_ids:
                await self.push_frame(frame, FrameDirection.UPSTREAM)
                seen_ids.add(frame.id)
            running = frame is not None
            self._up_queue.task_done()

    async def _process_down_queue(self):
        running = True
        seen_ids = set()
        while running:
            frame = await self._down_queue.get()
            if frame and frame.id not in seen_ids:
                await self.push_frame(frame, FrameDirection.DOWNSTREAM)
                seen_ids.add(frame.id)
            running = frame is not None
            self._down_queue.task_done()



from apipeline.processors.frame_processor import FrameProcessor


class NullFilter(FrameProcessor):
    """This filter doesn't allow passing any frames up or downstream."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



from asyncio import AbstractEventLoop
import logging
import os
import unittest

from apipeline.frames.control_frames import EndFrame
from apipeline.pipeline.pipeline import Pipeline, FrameDirection
from apipeline.pipeline.task import PipelineTask, PipelineParams
from apipeline.pipeline.runner import PipelineRunner
from apipeline.frames.data_frames import Frame, TextFrame, ImageRawFrame, AudioRawFrame
from apipeline.processors.filters.frame_filter import FrameFilter
from apipeline.processors.frame_processor import FrameProcessor
from apipeline.processors.logger import FrameLogger


"""
python -m unittest tests.filter.test_frame_filter.TestFilter
"""


class CheckFilter(FrameProcessor):
    def __init__(
        self,
        text: str | None = None,
        check_img: bool = False,
        *,
        name: str | None = None,
        loop: AbstractEventLoop | None = None,
        **kwargs,
    ):
        super().__init__(name=name, loop=loop, **kwargs)
        self._text = text
        self._check_img = check_img

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)
        if isinstance(frame, TextFrame):
            if self._text:
                assert frame.text != self._text, f"文本不应为:{self._text}"
        if self._check_img:
            assert not isinstance(frame, ImageRawFrame), "不应包含图像帧"
        await self.push_frame(frame, direction)


class TestFilter(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        logging.basicConfig(
            level=os.getenv("LOG_LEVEL", "info").upper(),
            format="%(asctime)s - %(name)s - %(levelname)s - %(pathname)s:%(lineno)d - %(funcName)s - %(message)s",
            handlers=[logging.StreamHandler()],
        )
        self.text = "你好"

        pipeline = Pipeline(
            [
                # FrameLogger(include_frame_types=[TextFrame]),
                FrameFilter(include_frame_types=[TextFrame]),
                # FrameLogger(),
                CheckFilter(text=self.text),
                FrameFilter(include_frame_types=[ImageRawFrame]),
                FrameLogger(),
                CheckFilter(text=self.text, check_img=True),
            ]
        )

        self.task = PipelineTask(pipeline, PipelineParams())

    async def asyncTearDown(self):
        pass

    async def test_run(self):
        runner = PipelineRunner()
        await self.task.queue_frame(TextFrame(self.text))
        await self.task.queue_frame(TextFrame("你好!"))
        await self.task.queue_frame(
            ImageRawFrame(image=bytes([]), size=(0, 0), format="PNG", mode="RGB")
        )
        await self.task.queue_frame(AudioRawFrame(audio=bytes([])))
        await self.task.queue_frame(EndFrame())
        await runner.run(self.task)



#!/usr/bin/env python
from setuptools import find_packages, setup

setup(
    name="apipeline",
    description="Python pipeline with asyncio support",
    long_description=open("README.md").read().strip(),
    long_description_content_type="text/markdown",
    version="0.2.5",
    license="BSD 3-Clause",
    url="https://github.com/weedge/pipeline-py",
    author="weedge",
    author_email="weege007@gmail.com",
    python_requires=">=3.10",
    packages=find_packages(
        include=[
            "apipeline.*",
        ],
        exclude=["*~"],
    ),
    include_package_data=True,
    install_requires=["pydantic >= 2.8.2"],
    project_urls={
        "Changes": "https://github.com/weedge/pipeline-py/releases",
        "Code": "https://github.com/weedge/pipeline-py",
        "Issue tracker": "https://github.com/weedge/pipeline-py/issues",
    },
    keywords=["pipeline", "asyncio"],
)



import logging
from abc import abstractmethod
import asyncio

from apipeline.frames.base import Frame
from apipeline.frames.sys_frames import CancelFrame, SystemFrame
from apipeline.frames.control_frames import EndFrame, StartFrame, ControlFrame
from apipeline.frames.data_frames import DataFrame
from apipeline.processors.async_frame_processor import AsyncFrameProcessor
from apipeline.processors.frame_processor import FrameDirection


class InputProcessor(AsyncFrameProcessor):
    """
    base inpurt processor
    """

    @abstractmethod
    async def start(self, frame: StartFrame):
        pass

    @abstractmethod
    async def stop(self):
        pass

    async def process_sys_frame(self, frame: Frame, direction: FrameDirection):
        logging.debug(
            f"f{self.__class__.__name__} process_sys_frame {frame} direction:{direction} doing"
        )
        await self.queue_frame(frame, direction)

    async def process_control_frame(self, frame: Frame, direction: FrameDirection):
        logging.debug(
            f"f{self.__class__.__name__} process_control_frame {frame} direction:{direction} doing"
        )
        await self.queue_frame(frame, direction)

    async def process_data_frame(self, frame: Frame, direction: FrameDirection):
        logging.debug(
            f"f{self.__class__.__name__} process_data_frame {frame} direction:{direction} doing"
        )
        await self.queue_frame(frame, direction)

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        # System frames
        if isinstance(frame, CancelFrame):
            # We don't queue a CancelFrame since we want to stop ASAP.
            await self.push_frame(frame, direction)
            await self.stop()
        # all other system frames
        elif isinstance(frame, SystemFrame):
            await self.process_sys_frame(frame, direction)
        # Control frames
        elif isinstance(frame, StartFrame):
            await self.queue_frame(frame, direction)
            await self.start(frame)
        elif isinstance(frame, EndFrame):
            await self.queue_frame(frame, direction)
            await self.stop()
        # all other control frames
        elif isinstance(frame, ControlFrame):
            await self.process_control_frame(frame, direction)
        # Data frames
        elif isinstance(frame, DataFrame):
            await self.process_data_frame(frame, direction)
        # Other frames
        else:
            await self.queue_frame(frame, direction)


class InputFrameProcessor(InputProcessor):
    """
    consume input asyncio.Queue, push to next processor;
    if input queue is None, use queue_frame method to push frame to next processor with direction

    WARNNING: dont to use it if use pipeline task, just use pipeline task queue_frame to input frame
    """

    def __init__(
        self,
        *,
        in_queue: asyncio.Queue | None = None,
        name: str | None = None,
        loop: asyncio.AbstractEventLoop | None = None,
        **kwargs,
    ):
        super().__init__(name=name, loop=loop, **kwargs)
        if in_queue is not None:
            self._in_queue = in_queue
            self._create_input_task()

    def _create_input_task(self):
        self._in_task = self.get_event_loop().create_task(self._in_push_task_handler())

    async def _in_push_task_handler(self):
        running = True
        while running:
            try:
                frame = await self._in_queue.get()
                await self.push_frame(frame)
                running = not isinstance(frame, EndFrame)
            except asyncio.CancelledError:
                break

    async def start(self, frame: Frame):
        if self._in_task is not None and self._in_task.cancelled():
            self._create_input_task()

    async def stop(self):
        if self._in_task is not None and self._in_task and not self._in_task.cancelled():
            self._in_task.cancel()
            await self._in_task

    async def cleanup(self):
        await self.stop()
        super().cleanup()



from typing import Tuple
from dataclasses import dataclass


from .base import Frame


@dataclass
class DataFrame(Frame):
    pass


@dataclass
class TextFrame(DataFrame):
    """A chunk of text. Emitted by LLM services, consumed by TTS services, can
    be used to send text through pipelines.

    """

    text: str

    def __str__(self):
        return f"{self.name}(text: {self.text})"


@dataclass
class AudioRawFrame(DataFrame):
    """A chunk of audio. Will be played by the transport if the transport's
    microphone has been enabled.

    """

    audio: bytes
    sample_rate: int = 16000
    num_channels: int = 1
    sample_width: int = 2

    def __post_init__(self):
        super().__post_init__()
        self.num_frames = int(len(self.audio) / (self.num_channels * self.sample_width))

    def __str__(self):
        return f"{self.name}(size: {len(self.audio)}, frames: {self.num_frames}, sample_rate: {self.sample_rate}, sample_width: {self.sample_width}, channels: {self.num_channels})"


@dataclass
class ImageRawFrame(DataFrame):
    """An image. Will be shown by the transport if the transport's camera is
    enabled.

    e.g.: image info from PIL.Image
    """

    image: bytes
    size: Tuple[int, int]
    format: str | None
    # https://pillow.readthedocs.io/en/stable/handbook/concepts.html#concept-modes
    mode: str | None

    def __str__(self):
        return f"{self.name}(size: {self.size}, format: {self.format}, mode: {self.mode})"



#
# Copyright (c) 2024, Daily
#
# SPDX-License-Identifier: BSD 2-Clause License
#

import asyncio
import logging
from typing import AsyncIterable, Iterable

from pydantic import BaseModel

from apipeline.frames.sys_frames import CancelFrame, ErrorFrame, Frame, MetricsFrame, StopTaskFrame
from apipeline.frames.control_frames import EndFrame, StartFrame
from apipeline.pipeline.base_pipeline import BasePipeline
from apipeline.processors.frame_processor import FrameDirection, FrameProcessor
from apipeline.utils.obj import obj_count, obj_id


class PipelineParams(BaseModel):
    allow_interruptions: bool = False
    enable_metrics: bool = False
    enable_usage_metrics: bool = False
    send_initial_empty_metrics: bool = True
    report_only_initial_ttfb: bool = False


class Source(FrameProcessor):
    def __init__(self, up_queue: asyncio.Queue):
        super().__init__()
        self._up_queue = up_queue

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        match direction:
            case FrameDirection.UPSTREAM:
                await self._handle_upstream_frame(frame)
            case FrameDirection.DOWNSTREAM:
                await self.push_frame(frame, direction)

    async def _handle_upstream_frame(self, frame: Frame):
        if isinstance(frame, ErrorFrame):
            logging.error(f"Error running app: {frame}")
            if frame.fatal:
                # Cancel all tasks downstream.
                await self.push_frame(CancelFrame())
                # Tell the task we should stop.
                await self._up_queue.put(StopTaskFrame())


class PipelineTask:
    def __init__(self, pipeline: BasePipeline, params: PipelineParams = PipelineParams()):
        self.id: int = obj_id()
        self.name: str = f"{self.__class__.__name__}#{obj_count(self)}"

        self._pipeline = pipeline
        self._params = params
        self._finished = False

        self._down_queue = asyncio.Queue()
        self._up_queue = asyncio.Queue()

        self._source = Source(self._up_queue)
        self._source.link(pipeline)

    def has_finished(self):
        return self._finished

    async def stop_when_done(self):
        logging.debug(f"Task {self} scheduled to stop when done")
        await self.queue_frame(EndFrame())

    async def cancel(self):
        logging.debug(f"Canceling pipeline task {self}")
        # Make sure everything is cleaned up downstream. This is sent
        # out-of-band from the main streaming task which is what we want since
        # we want to cancel right away.
        await self._source.push_frame(CancelFrame())
        self._process_down_task.cancel()
        self._process_up_task.cancel()
        await self._process_down_task
        await self._process_up_task

    async def run(self):
        self._process_up_task = asyncio.create_task(self._process_up_queue())
        self._process_down_task = asyncio.create_task(self._process_down_queue())
        await asyncio.gather(self._process_up_task, self._process_down_task)
        self._finished = True

    async def queue_frame(self, frame: Frame):
        await self._down_queue.put(frame)

    async def queue_frames(self, frames: Iterable[Frame] | AsyncIterable[Frame]):
        if isinstance(frames, AsyncIterable):
            async for frame in frames:
                await self.queue_frame(frame)
        elif isinstance(frames, Iterable):
            for frame in frames:
                await self.queue_frame(frame)
        else:
            raise Exception("Frames must be an iterable or async iterable")

    def _initial_metrics_frame(self) -> MetricsFrame:
        processors = self._pipeline.processors_with_metrics()
        ttfb = [{"processor": p.name, "value": 0.0} for p in processors]
        processing = [{"processor": p.name, "value": 0.0} for p in processors]
        return MetricsFrame(ttfb=ttfb, processing=processing)

    async def _process_down_queue(self):
        start_frame = StartFrame(
            allow_interruptions=self._params.allow_interruptions,
            enable_metrics=self._params.enable_metrics,
            enable_usage_metrics=self._params.enable_usage_metrics,
            report_only_initial_ttfb=self._params.report_only_initial_ttfb,
        )
        await self._source.process_frame(start_frame, FrameDirection.DOWNSTREAM)
        if self._params.enable_metrics and self._params.send_initial_empty_metrics:
            await self._source.process_frame(
                self._initial_metrics_frame(), FrameDirection.DOWNSTREAM
            )

        running = True
        should_cleanup = True
        while running:
            try:
                frame = await self._down_queue.get()
                await self._source.process_frame(frame, FrameDirection.DOWNSTREAM)
                running = not (isinstance(frame, StopTaskFrame) or isinstance(frame, EndFrame))
                should_cleanup = not isinstance(frame, StopTaskFrame)
                self._down_queue.task_done()
            except asyncio.CancelledError:
                break
        # Cleanup only if we need to.
        if should_cleanup:
            await self._source.cleanup()
            await self._pipeline.cleanup()
        # We just enqueue None to terminate the task gracefully.
        self._process_up_task.cancel()
        await self._process_up_task

    async def _process_up_queue(self):
        while True:
            try:
                frame = await self._up_queue.get()
                if isinstance(frame, StopTaskFrame):
                    await self.queue_frame(StopTaskFrame())
                self._up_queue.task_done()
            except asyncio.CancelledError:
                break

    def __str__(self):
        return self.name
